# Copyright (c) 2016 iXsystems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Contains configuration options for iXsystems Cinder drivers."""


from oslo_config import cfg


ixsystems_connection_opts = [
    cfg.StrOpt('ixsystems_server_hostname',
               default=None,
               help='Host name for the storage controller'),
    cfg.IntOpt('ixsystems_server_port',
               default=3000,
               help='Port number for the storage controller'),
    cfg.IntOpt('ixsystems_server_iscsi_port',
               default=3260,
               help='ISCSI port number for the storage controller'),
    cfg.StrOpt('ixsystems_api_version',
               default='v2.0',
               help='FREENAS API version'),
    cfg.StrOpt('ixsystems_volume_backend_name',
               default='iXsystems_FREENAS_Storage',
               help='Backend Storage Controller Name'),
    cfg.StrOpt('ixsystems_vendor_name',
               default='iXsystem',
               help='vendor name on Storage controller'),
    cfg.StrOpt('ixsystems_storage_protocol',
               default='iscsi',
               help='storage protocol on Storage controller'), ]

ixsystems_transport_opts = [
    cfg.StrOpt('ixsystems_transport_type',
               default='http',
               help='Transport type protocol'), ]

ixsystems_basicauth_opts = [
    cfg.StrOpt('ixsystems_login',
               default='root',
               help='User name for the storage controller'),
    cfg.StrOpt('ixsystems_password',
               default='ixsystems',
               help='Password for the storage controller',
               secret=True), ]

ixsystems_provisioning_opts = [
    cfg.StrOpt('ixsystems_datastore_pool',
               default=None,
               help='Storage controller datastore pool name'),
    cfg.StrOpt('ixsystems_dataset_path',
               default=None,
               help='Datasset full path name including pool'),
    cfg.IntOpt('ixsystems_reserved_percentage',
               default=0,
               help='Reserved space on Storage controller'),
    cfg.StrOpt('ixsystems_iqn_prefix',
               default='iqn.2005-10.org.freenas.ctl',
               help='Storage controller iSCSI Qualified Name prefix'),
    cfg.StrOpt('ixsystems_portal_id',
               default=1,
               help='Storage controller iSCSI portal ID'),
    cfg.StrOpt('ixsystems_initiator_id',
               default=1,
               help='Storage controller iSCSI Initiator ID'), ]



# Copyright (c) 2016 iXsystems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


def get_size_in_gb(size_in_bytes):
    """convert size in gbs"""
    return size_in_bytes / (1024 * 1024 * 1024)


def get_bytes_from_gb(size_in_gb):
    """Convert size from GB into bytes."""
    return size_in_gb * (1024 * 1024 * 1024)


def generate_freenas_volume_name(name, iqn_prefix):
    """Create FREENAS volume / iscsitarget name from Cinder name."""
    backend_volume = 'volume-' + name.split('-')[1]
    backend_target = 'target-' + name.split('-')[1]
    backend_iqn = iqn_prefix + backend_target
    return {'name': backend_volume,
            'target': backend_target, 'iqn': backend_iqn}


def generate_freenas_snapshot_name(name, iqn_prefix):
    """Create FREENAS snapshot / iscsitarget name from Cinder name."""
    backend_snap = 'snap-' + name.split('-')[1]
    backend_target = 'target-' + name.split('-')[1]
    backend_iqn = iqn_prefix + backend_target
    return {'name': backend_snap,
            'target': backend_target, 'iqn': backend_iqn}


def get_iscsi_portal(hostname, port):
    """Get iscsi portal info from iXsystems FREENAS configuration."""
    return "%s:%s" % (hostname, port)



# Copyright (c) 2016 iXsystems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
FREENAS api for iXsystems FREENAS system.

Contains classes required to issue REST based api calls to FREENAS system.
"""

import base64
import simplejson as json

from oslo_log import log as logging
import urllib.error
import urllib.parse
import urllib.request

LOG = logging.getLogger(__name__)


class FreeNASServer(object):
    """FreeNAS server connection logic."""

    FREENAS_API_VERSION = "v2.0"
    TRANSPORT_TYPE = 'http'
    STYLE_LOGIN_PASSWORD = 'basic_auth'

    # FREENAS Commands
    SELECT_COMMAND = 'select'
    CREATE_COMMAND = 'create'
    UPDATE_COMMAND = 'update'
    DELETE_COMMAND = 'delete'

    # FREENAS API query tables
    REST_API_VOLUME = "/pool/dataset"
    REST_API_EXTENT = "/iscsi/extent"
    REST_API_TARGET = "/iscsi/target"
    REST_API_TARGET_TO_EXTENT = "/iscsi/targetextent"
    # REST_API_TARGET_GROUP = "/services/iscsi/targetgroup/"
    REST_API_SNAPSHOT = "/zfs/snapshot"
    ZVOLS = "zvols"
    TARGET_ID = -1  # We assume invalid id to begin with
    STORAGE_TABLE = "/storage"
    CLONE = "clone"
    # Command response format
    COMMAND_RESPONSE = {'status': '%s',
                        'response': '%s'}

    # Command status
    STATUS_OK = 'ok'
    STATUS_ERROR = 'error'

    def __init__(self, host, port,
                 username=None, password=None,
                 api_version=FREENAS_API_VERSION,
                 transport_type=TRANSPORT_TYPE,
                 style=STYLE_LOGIN_PASSWORD):
        self._host = host
        self._port = port
        self._username = username
        self._password = password
        self.set_api_version(api_version)
        self.set_transport_type(transport_type)
        self.set_style(style)

    def get_host(self):
        return self._host

    def get_port(self):
        return self._port

    def get_username(self):
        return self._username

    def get_password(self):
        return self._password

    def get_transport_type(self):
        return self._protocol

    def set_host(self, host):
        self._host = host

    def set_port(self, port):
        try:
            int(port)
        except ValueError:
            raise ValueError("Port must be integer")

    def set_username(self, username):
        self._username = username

    def set_password(self, password):
        self._password = password

    def set_api_version(self, api_version):
        self._api_version = api_version

    def set_transport_type(self, transport_type):
        self._protocol = transport_type

    def set_style(self, style):
        """Set the authorization style for communicating with the server.

           Supports basic_auth for now.
        """
        if style.lower() not in (FreeNASServer.STYLE_LOGIN_PASSWORD):
            raise ValueError('Unsupported authentication style')
        self._auth_style = style.lower()

    def get_url(self):
        """Returns connection string.

           built using _protocol, _host, _port and _api_version fields
        """
        return '%s://%s/api/%s' % (self._protocol,
                                   self._host,
                                   self._api_version)

    def _create_request(self, request_d, param_list):
        """Creates urllib2.Request object."""
        if not self._username or not self._password:
            raise ValueError("Invalid username/password combination")
        loginstring = ("%s:%s" % (self._username, self._password))
        bloginstring = bytes(loginstring, encoding='utf8')
        bauth = base64.b64encode(bloginstring)
        auth = bauth.decode("utf8")
        headers = {'Content-Type': 'application/json',
                   'Authorization': 'Basic %s' % (auth,)}
        url = self.get_url() + request_d
        LOG.debug('url : %s, request: %s', url, request_d)
        LOG.debug('param list : %s', param_list)
        return urllib.request.Request(url, param_list, headers)

    def _get_method(self, command_d):
        """Select http method based on FREENAS command."""
        if command_d == self.SELECT_COMMAND:
            return 'GET'
        elif command_d == self.CREATE_COMMAND:
            return 'POST'
        elif command_d == self.DELETE_COMMAND:
            return 'DELETE'
        elif command_d == self.UPDATE_COMMAND:
            return 'PUT'
        else:
            return None

    def _parse_result(self, command_d, response_d):
        """parses the response upon execution of FREENAS API.

           COMMAND_RESPONSE is
           the dictionary object with status and response fields.
           If error, set status to ERROR else set it to OK
        """
        response_str = response_d.read()
        status = None
        if command_d == self.SELECT_COMMAND:
            status = self.STATUS_OK
            response_obj = response_str
        elif (command_d == self.CREATE_COMMAND or
              command_d == self.DELETE_COMMAND or
              command_d == self.UPDATE_COMMAND):
            response_obj = response_str
            status = self.STATUS_OK
        else:
            status = self.STATUS_ERROR
            response_obj = None

        self.COMMAND_RESPONSE['status'] = status
        self.COMMAND_RESPONSE['response'] = response_obj
        return self.COMMAND_RESPONSE

    def _get_error_info(self, err):
        """Collects error response message."""
        self.COMMAND_RESPONSE['status'] = self.STATUS_ERROR
        if isinstance(err, urllib.error.HTTPError):
            self.COMMAND_RESPONSE['response'] = '%d:%s' % (err.code, err.msg)
        elif isinstance(err, urllib.error.URLError):
            self.COMMAND_RESPONSE['response'] = '%s:%s' % \
                                                (str(err.reason.errno),
                                                 err.reason.strerror)
        else:
            return None
        return self.COMMAND_RESPONSE

    def invoke_command(self, command_d, request_d, param_list):
        """Invokes api and returns response object."""
        LOG.debug('invoke_command')
        request = self._create_request(request_d, param_list)
        method = self._get_method(command_d)
        if not method:
            raise FreeNASApiError("Invalid FREENAS command")
        request.get_method = lambda: method
        try:
            response_d = urllib.request.urlopen(request)
            response = self._parse_result(command_d, response_d)
            LOG.debug("invoke_command : response for request %s : %s",
                      request_d, json.dumps(response))
        except urllib.error.HTTPError as e:
            # LOG the error message received from FreeNAS/TrueNAS:
            # https://github.com/iXsystems/cinder/issues/11
            LOG.info('Error returned from server: "%s"',
                     json.loads(e.read().decode('utf8'))['message'])
            error_d = self._get_error_info(e)
            if error_d:
                return error_d
            else:
                raise FreeNASApiError(e.code, e.msg)
        except Exception as e:
            error_d = self._get_error_info(e)
            if error_d:
                return error_d
            else:
                raise FreeNASApiError('Unexpected error', e)
        return response


class FreeNASApiError(Exception):
    """Base exception class for FREENAS api errors."""

    def __init__(self, code='unknown', message='unknown'):
        self.code = code
        self.message = message

    def __str__(self, *args, **kwargs):
        return 'FREENAS api failed. Reason - %s:%s' % (self.code, self.message)



# Copyright (c) 2016 iXsystems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import simplejson as json
import urllib.parse

from cinder import exception
from cinder.i18n import _
from driver.freenasapi import FreeNASApiError
from driver.freenasapi import FreeNASServer
from driver import utils as ix_utils
from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class TrueNASCommon(object):

    VERSION = "2.0.0"
    IGROUP_PREFIX = 'openstack-'

    required_flags = ['ixsystems_transport_type', 'ixsystems_login',
                      'ixsystems_password', 'ixsystems_server_hostname',
                      'ixsystems_server_port', 'ixsystems_server_iscsi_port',
                      'ixsystems_volume_backend_name', 'ixsystems_vendor_name',
                      'ixsystems_storage_protocol', 'ixsystems_datastore_pool',
                      'ixsystems_dataset_path', 'ixsystems_iqn_prefix', ]

    def __init__(self, configuration=None):
        self.configuration = configuration
        self.backend_name = self.configuration.ixsystems_volume_backend_name
        self.vendor_name = self.configuration.ixsystems_vendor_name
        self.storage_protocol = self.configuration.ixsystems_storage_protocol
        self.stats = {}

    def _create_handle(self, **kwargs):
        """Instantiate client for API comms with iXsystems FREENAS server."""

        host_system = kwargs['hostname']
        LOG.debug('Using iXsystems FREENAS server: %s', host_system)
        self.handle = FreeNASServer(
            host=host_system,
            port=kwargs['port'],
            username=kwargs['login'],
            password=kwargs['password'],
            api_version=kwargs['api_version'],
            transport_type=kwargs['transport_type'],
            style=FreeNASServer.STYLE_LOGIN_PASSWORD)
        if not self.handle:
            raise FreeNASApiError("Failed to create handle for FREENAS server")

    def _check_flags(self):
        """Check if any required iXsystems configuration flag is missing."""
        for flag in self.required_flags:
            if not getattr(self.configuration, flag, None):
                raise exception.CinderException(_('%s is not set') % flag)

    def _do_custom_setup(self):
        """Setup iXsystems FREENAS driver."""
        self._create_handle(
            hostname=self.configuration.ixsystems_server_hostname,
            port=self.configuration.ixsystems_server_port,
            login=self.configuration.ixsystems_login,
            password=self.configuration.ixsystems_password,
            api_version=self.configuration.ixsystems_api_version,
            transport_type=self.configuration.ixsystems_transport_type)

        if not self.handle:
            raise FreeNASApiError(
                "Failed to create handle for FREENAS server")

    def _create_volume(self, name, size):
        """Creates a volume of specified size."""

        params = {}
        params['name'] = self.configuration.ixsystems_dataset_path + '/' + name
        params['type'] = 'VOLUME'
        params['volsize'] = ix_utils.get_bytes_from_gb(size)
        jparams = json.dumps(params)
        jparams = jparams.encode('utf8')
        request_urn = ('%s') % (FreeNASServer.REST_API_VOLUME)
        LOG.debug('_create_volume params : %s', params)
        LOG.debug('_create_volume urn : %s', request_urn)
        ret = self.handle.invoke_command(FreeNASServer.CREATE_COMMAND,
                                         request_urn, jparams)
        LOG.debug('_create_volume response : %s', json.dumps(ret))
        if ret['status'] != FreeNASServer.STATUS_OK:
            msg = ('Error while creating volume: %s' % ret['response'])
            raise FreeNASApiError('Unexpected error', msg)

    def _target_to_extent(self, target_id, extent_id):
        """Create relationship between iscsi target to iscsi extent."""

        LOG.debug('_target_to_extent target id : %s extend id : %s',
                  target_id, extent_id)

        request_urn = ('%s/') % (FreeNASServer.REST_API_TARGET_TO_EXTENT)
        params = {}
        params['target'] = target_id
        params['extent'] = extent_id
        # params['iscsi_lunid'] = 0   # no longer needed with API v2.0
        jparams = json.dumps(params)
        jparams = jparams.encode('utf8')

        LOG.debug('_create_target_to_extent params : %s', json.dumps(params))

        tgt_ext = self.handle.invoke_command(FreeNASServer.CREATE_COMMAND,
                                             request_urn, jparams)

        LOG.debug('_target_to_extent response : %s', json.dumps(tgt_ext))

        if tgt_ext['status'] != FreeNASServer.STATUS_OK:
            msg = ('Error while creating relation between '
                   'target and extent: %s' % tgt_ext['response'])
            raise FreeNASApiError('Unexpected error', msg)

    def _create_target(self, name):
        # v2.0 API - targetgroup can now be added when target is created
        targetgroup_params = [{}]
        # TODO: Decide to create portal or not
        targetgroup_params[0]['portal'] = int(
            self.configuration.ixsystems_portal_id)

        # TODO: Decide to create initiator or not
        targetgroup_params[0]['initiator'] = int(
            self.configuration.ixsystems_initiator_id)
        tgt_params = {}
        tgt_params['name'] = name
        tgt_params['groups'] = targetgroup_params
        jtgt_params = json.dumps(tgt_params)
        jtgt_params = jtgt_params.encode('utf8')
        LOG.debug('_create_target params : %s', json.dumps(tgt_params))
        request_urn = ('%s/') % (FreeNASServer.REST_API_TARGET)
        target = self.handle.invoke_command(FreeNASServer.CREATE_COMMAND,
                                            request_urn, jtgt_params)
        LOG.debug('_create_target response : %s', json.dumps(target))

        if target['status'] != FreeNASServer.STATUS_OK:
            msg = ('Error while creating iscsi target: {}'.format(
                target['response']))
            raise FreeNASApiError('Unexpected error', msg)

        target_id = json.loads(target['response'])['id']
        # self._create_target_group(target_id)

        return target_id

    def _create_extent(self, name, volume_name, from_snapshot=False):
        ext_params = {}
        if from_snapshot:
            ext_params['Source'] = volume_name
        else:
            ext_params['type'] = 'DISK'
            ext_params['name'] = name
        ext_params['disk'] = ('zvol/%s/%s') % (
            self.configuration.ixsystems_dataset_path, volume_name)
        jext_params = json.dumps(ext_params)
        LOG.debug('_create_extent params : %s', jext_params)
        jext_params = jext_params.encode('utf8')
        request_urn = ('%s/') % (FreeNASServer.REST_API_EXTENT)
        extent = self.handle.invoke_command(FreeNASServer.CREATE_COMMAND,
                                            request_urn, jext_params)

        LOG.debug('_create_extent response : %s', json.dumps(extent))

        if extent['status'] != FreeNASServer.STATUS_OK:
            msg = ('Error while creating iscsi target extent: {}'.format(
                extent['response']))
            raise FreeNASApiError('Unexpected error', msg)

        return json.loads(extent['response'])['id']

    def get_iscsitarget_id(self, name):
        """get iscsi target id from target name."""

        request_urn = ('%s') % (FreeNASServer.REST_API_TARGET)
        LOG.debug('get_iscsitarget_id request_urn : %s', request_urn)
        ret = self.handle.invoke_command(FreeNASServer.SELECT_COMMAND,
                                         request_urn, None)
        LOG.debug('get_iscsitarget_id response : %s', json.dumps(ret))
        if ret['status'] != FreeNASServer.STATUS_OK:
            msg = ('Error while deleting iscsi target: %s' % ret['response'])
            raise FreeNASApiError('Unexpected error', msg)

        uresp = ret['response']
        resp = json.loads(uresp.decode('utf8'))
        try:
            return (item for item in resp
                    if item['name'] == name).__next__()['id']
        except StopIteration:
            return 0

    def get_tgt_ext_id(self, name):
        """Get target-extent mapping id from target name."""

        request_urn = ('%s') % (FreeNASServer.REST_API_TARGET_TO_EXTENT)
        LOG.debug('get_tgt_ext_id urn : %s', request_urn)
        ret = self.handle.invoke_command(FreeNASServer.SELECT_COMMAND,
                                         request_urn, None)
        LOG.debug('get_tgt_ext_id response : %s', json.dumps(ret))
        if ret['status'] != FreeNASServer.STATUS_OK:
            msg = ('Error while deleting iscsi target: %s' % ret['response'])
            raise FreeNASApiError('Unexpected error', msg)

        uresp = ret['response']
        resp = json.loads(uresp.decode('utf8'))
        try:
            return (item for item in resp
                    if item['target'] == name).__next__()['id']
        except StopIteration:
            return 0

    def get_extent_id(self, name):
        """Get Extent ID from Extent Name."""

        request_urn = ('%s') % (FreeNASServer.REST_API_EXTENT)
        LOG.debug('get_extent_id urn : %s', request_urn)
        ret = self.handle.invoke_command(FreeNASServer.SELECT_COMMAND,
                                         request_urn, None)
        LOG.debug('get_extent_id response : %s', json.dumps(ret))
        if ret['status'] != FreeNASServer.STATUS_OK:
            msg = ('Error while getting extent id: %s' % ret['response'])
            raise FreeNASApiError('Unexpected error', msg)

        uresp = ret['response']
        resp = json.loads(uresp.decode('utf8'))
        try:
            return (item for item in resp
                    if item['name'] == name).__next__()['id']
        except StopIteration:
            return 0

    def _create_iscsitarget(self, name, volume_name):
        """Creates a iSCSI target on specified volume OR snapshot.

           TODO : Skipped part for snapshot, review once iscsi target working
           TODO: Add cleanup if any operation fails
        """

        # Create iscsi target for specified volume
        tgt_id = self._create_target(name)

        # Create extent for iscsi target for specified volume
        ext_id = self._create_extent(name, volume_name)

        # Create target to extent mapping for specified volume
        self._target_to_extent(tgt_id, ext_id)

    def delete_target_to_extent(self, tgt_ext_id):
        pass

    def delete_target(self, target_id):
        if target_id:
            request_urn = ('%s/id/%s') % (
                FreeNASServer.REST_API_TARGET, target_id)
            LOG.debug('delete_target urn : %s', request_urn)
            ret = self.handle.invoke_command(FreeNASServer.DELETE_COMMAND,
                                             request_urn, None)
            LOG.debug('delete_target response : %s', json.dumps(ret))
            if ret['status'] != FreeNASServer.STATUS_OK:
                msg = (
                    'Error while deleting iscsi target: %s' % ret['response'])
                raise FreeNASApiError('Unexpected error', msg)

    def delete_extent(self, extent_id):
        if extent_id:
            request_urn = ('%s/id/%s') % (
                FreeNASServer.REST_API_EXTENT, extent_id)
            LOG.debug('delete_extent urn : %s', request_urn)
            ret = self.handle.invoke_command(FreeNASServer.DELETE_COMMAND,
                                             request_urn, None)
            LOG.debug('delete_extent response : %s', json.dumps(ret))
            if ret['status'] != FreeNASServer.STATUS_OK:
                msg = (
                    'Error while deleting iscsi extent: %s' % ret['response'])
                raise FreeNASApiError('Unexpected error', msg)

    def _delete_iscsitarget(self, name):
        """Deletes specified iSCSI target."""
        tgt_ext_id = self.get_tgt_ext_id(name)
        target_id = self.get_iscsitarget_id(name)
        extent_id = self.get_extent_id(name)

        self.delete_target_to_extent(tgt_ext_id)
        self.delete_target(target_id)
        self.delete_extent(extent_id)

    def _dependent_clone(self, name):
        """returns the fullname of snapshot used to create volume 'name'."""
        request_urn = ('%s/id/%s%s') % (
            FreeNASServer.REST_API_VOLUME,
            urllib.parse.quote_plus(
                self.configuration.ixsystems_dataset_path + '/'),
            name)
        LOG.debug('_dependent_clones urn : %s', request_urn)
        ret = self.handle.invoke_command(FreeNASServer.SELECT_COMMAND,
                                         request_urn, None)
        LOG.debug('_dependent_clones response : %s', json.dumps(ret))
        if ret['status'] != FreeNASServer.STATUS_OK:
            msg = ('Error while getting volume: %s' % ret['response'])
            raise FreeNASApiError('Unexpected error', msg)
        uresp = ret['response']
        resp = json.loads(uresp.decode('utf8'))
        return resp['origin']['value']

    def _delete_volume(self, name):
        """Deletes specified volume."""
        request_urn = ('%s/id/%s%s') % (
            FreeNASServer.REST_API_VOLUME,
            urllib.parse.quote_plus(
                self.configuration.ixsystems_dataset_path + '/'),
            name)
        LOG.debug('_delete_volume urn : %s', request_urn)
        # add check for dependent clone, if exists will delete
        clone = self._dependent_clone(name)
        ret = self.handle.invoke_command(FreeNASServer.DELETE_COMMAND,
                                         request_urn, None)
        LOG.debug('_delete_volume response : %s', json.dumps(ret))

        # delete the cloned-from snapshot.
        # Must check before deleting volume, but delete snapshot after
        if clone:
            fullvolume, snapname = clone.split('@')
            temp, snapvol = fullvolume.rsplit('/', 1)
            self._delete_snapshot(snapname, snapvol)
        if ret['status'] != FreeNASServer.STATUS_OK:
            msg = ('Error while deleting volume: %s' % ret['response'])
            raise FreeNASApiError('Unexpected error', msg)

    def _create_snapshot(self, name, volume_name):
        """Creates a snapshot of specified volume."""
        args = {}
        args['dataset'] = ('%s/%s') % (
            self.configuration.ixsystems_dataset_path, volume_name)
        args['name'] = name
        jargs = json.dumps(args)
        jargs = jargs.encode("utf8")
        request_urn = ('%s') % (FreeNASServer.REST_API_SNAPSHOT)
        LOG.debug('_create_snapshot urn : %s', request_urn)

        try:
            ret = self.handle.invoke_command(FreeNASServer.CREATE_COMMAND,
                                             request_urn, jargs)
            LOG.debug('_create_snapshot response : %s', json.dumps(ret))
            if ret['status'] != FreeNASServer.STATUS_OK:
                msg = ('Error while creating snapshot: %s' % ret['response'])
                raise FreeNASApiError('Unexpected error', msg)
        except Exception as e:
            raise FreeNASApiError('Unexpected error', e)

    def _delete_snapshot(self, name, volume_name):
        """Delets a snapshot of specified volume."""
        LOG.debug('_delete_snapshot, deleting name: %s from volume: %s',
                  name, volume_name)
        request_urn = ('%s/id/%s@%s') % (
            FreeNASServer.REST_API_SNAPSHOT,
            urllib.parse.quote_plus(
                self.configuration.ixsystems_dataset_path + '/' + volume_name),
            name)
        LOG.debug('_delete_snapshot urn : %s', request_urn)
        try:
            ret = self.handle.invoke_command(FreeNASServer.DELETE_COMMAND,
                                             request_urn, None)
            LOG.debug('_delete_snapshot response : %s', json.dumps(ret))
            if ret['status'] != FreeNASServer.STATUS_OK:
                msg = ('Error while deleting snapshot: %s' % ret['response'])
                raise FreeNASApiError('Unexpected error', msg)
        except Exception as e:
            raise FreeNASApiError('Unexpected error', e)

    def _create_volume_from_snapshot(self, name, snapshot_name,
                                     snap_zvol_name):
        """creates a volume from a snapshot"""
        args = {}
        args['snapshot'] = ('%s/%s@%s') % (
            self.configuration.ixsystems_dataset_path,
            snap_zvol_name,
            snapshot_name)
        args['dataset_dst'] = ('%s/%s') % (
            self.configuration.ixsystems_dataset_path, name)
        jargs = json.dumps(args)
        jargs = jargs.encode("utf8")
        request_urn = ('%s/%s') % (
            FreeNASServer.REST_API_SNAPSHOT, FreeNASServer.CLONE)
        LOG.debug('_create_volume_from_snapshot urn : %s', request_urn)
        try:
            ret = self.handle.invoke_command(FreeNASServer.CREATE_COMMAND,
                                             request_urn, jargs)
            LOG.debug('_create_volume_from_snapshot response : %s',
                      json.dumps(ret))
            if ret['status'] != FreeNASServer.STATUS_OK:
                msg = ('Error while creating snapshot: %s' % ret['response'])
                raise FreeNASApiError('Unexpected error', msg)
        except Exception as e:
            raise FreeNASApiError('Unexpected error', e)

    def _update_volume_stats(self):
        """Retrieve stats info from volume group.

           REST API: $ GET /pools/mypool "size":95,"allocated":85,
        """
        # HACK: for now, use an API v1.0 call to get
        # these stats until available in v2.0 API
        self.handle.set_api_version('v1.0')
        request_urn = ('%s/%s/') % (
            '/storage/volume',
            self.configuration.ixsystems_datastore_pool)
        LOG.debug('_update_volume_stats request_urn : %s', request_urn)
        ret = self.handle.invoke_command(FreeNASServer.SELECT_COMMAND,
                                         request_urn, None)
        LOG.debug("_update_volume_stats response : %s", json.dumps(ret))
        data = {}
        data["volume_backend_name"] = self.backend_name
        data["vendor_name"] = self.vendor_name
        data["driver_version"] = self.VERSION
        data["storage_protocol"] = self.storage_protocol
        data['total_capacity_gb'] = ix_utils.get_size_in_gb(
            json.loads(ret['response'])['avail'] +
            json.loads(ret['response'])['used'])
        data['free_capacity_gb'] = ix_utils.get_size_in_gb(
            json.loads(ret['response'])['avail'])
        data['reserved_percentage'] = (
            self.configuration.ixsystems_reserved_percentage)
        data['reserved_percentage'] = 0
        data['QoS_support'] = False

        self.stats = data
        # set back to v2.0 api for other calls...
        self.handle.set_api_version('v2.0')
        return self.stats

    def _create_cloned_volume_to_snapshot_map(self, volume_name, snapshot):
        """maintain a mapping between cloned volume and tempary snapshot."""
        map_file = os.path.join(CONF.volumes_dir, volume_name)
        jparams = json.dumps(snapshot)
        try:
            fd = open(map_file, 'w+')
            fd.write(jparams)
            fd.close()
        except Exception as e:
            LOG.error('_create_halo_volume_name_map: %s', e)

    def _extend_volume(self, name, new_size):
        """Extend an existing volumes size."""
        LOG.debug('_extend__volume name: %s', name)
        params = {}
        params['volsize'] = ix_utils.get_bytes_from_gb(new_size)
        jparams = json.dumps(params)
        jparams = jparams.encode('utf8')
        request_urn = ('%s/id/%s') % (
            FreeNASServer.REST_API_VOLUME,
            urllib.parse.quote_plus(
                self.configuration.ixsystems_dataset_path + '/' + name))
        ret = self.handle.invoke_command(FreeNASServer.UPDATE_COMMAND,
                                         request_urn, jparams)
        if ret['status'] != FreeNASServer.STATUS_OK:
            msg = ('Error while extending volume: %s' % ret['response'])
            raise FreeNASApiError('Unexpected error', msg)

    def _create_export(self, volume_name):
        freenas_volume = ix_utils.generate_freenas_volume_name(
            volume_name,
            self.configuration.ixsystems_iqn_prefix)

        if freenas_volume is None:
            LOG.error('Error in exporting FREENAS volume!')
            handle = None
        else:
            handle = "%s:%s,%s %s" % \
                     (self.configuration.ixsystems_server_hostname,
                      self.configuration.ixsystems_server_iscsi_port,
                      freenas_volume['target'],
                      freenas_volume['iqn'])

        LOG.debug('provider_location: %s', handle)
        return handle



# Copyright (c) 2016 iXsystems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Volume driver for iXsystems iSCSI storage systems.

This driver requires iXsystems storage systems with installed iSCSI licenses.
"""

import simplejson as json

from cinder.volume import driver
from driver import common
from driver.options import ixsystems_basicauth_opts
from driver.options import ixsystems_connection_opts
from driver.options import ixsystems_provisioning_opts
from driver.options import ixsystems_transport_opts
from driver import utils as ix_utils
from oslo_config import cfg
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
CONF.register_opts(ixsystems_connection_opts)
CONF.register_opts(ixsystems_transport_opts)
CONF.register_opts(ixsystems_basicauth_opts)
CONF.register_opts(ixsystems_provisioning_opts)


class FreeNASISCSIDriver(driver.ISCSIDriver):
    """FREENAS iSCSI volume driver."""

    VERSION = "2.0.0"
    IGROUP_PREFIX = 'openstack-'

    required_flags = ['ixsystems_transport_type', 'ixsystems_login',
                      'ixsystems_password', 'ixsystems_server_hostname',
                      'ixsystems_server_port', 'ixsystems_server_iscsi_port',
                      'ixsystems_volume_backend_name', 'ixsystems_vendor_name',
                      'ixsystems_storage_protocol', 'ixsystems_datastore_pool',
                      'ixsystems_dataset_path', 'ixsystems_iqn_prefix', ]

    def __init__(self, *args, **kwargs):
        """Initialize FreeNASISCSIDriver Class."""

        LOG.info('iXsystems: Init Cinder Driver')
        super(FreeNASISCSIDriver, self).__init__(*args, **kwargs)
        self.configuration.append_config_values(ixsystems_connection_opts)
        self.configuration.append_config_values(ixsystems_basicauth_opts)
        self.configuration.append_config_values(ixsystems_transport_opts)
        self.configuration.append_config_values(ixsystems_provisioning_opts)
        self.configuration.ixsystems_iqn_prefix += ':'
        self.common = common.TrueNASCommon(configuration=self.configuration)
        self.stats = {}

    def check_for_setup_error(self):
        """Check for iXsystems FREENAS configuration parameters."""
        LOG.info('iXSystems: Check For Setup Error')
        self.common._check_flags()

    def do_setup(self, context):
        """Setup iXsystems FREENAS driver.

            Check for configuration flags and setup iXsystems FREENAS client
        """
        LOG.info('iXsystems Do Setup')
        # TODO:add check to see if volume exist, able to connect
        # truenas array
        self.check_for_setup_error()
        self.common._do_custom_setup()

    def create_volume(self, volume):
        """Creates a volume of specified size and export it as iscsi target."""
        LOG.info('iXsystems Create Volume')
        LOG.debug('create_volume : volume name :: %s', volume['name'])

        freenas_volume = ix_utils.generate_freenas_volume_name(
            volume['name'],
            self.configuration.ixsystems_iqn_prefix)

        LOG.debug('volume name after freenas generate : %s',
                  json.dumps(freenas_volume))

        freenas_volume['size'] = volume['size']
        freenas_volume['target_size'] = volume['size']

        self.common._create_volume(freenas_volume['name'],
                                   freenas_volume['size'])
        # Remove LUN Creation from here,check at initi
        self.common._create_iscsitarget(freenas_volume['target'],
                                        freenas_volume['name'])

    def delete_volume(self, volume):
        """Deletes volume and corresponding iscsi target."""
        LOG.info('iXsystems Delete Volume')
        LOG.debug('delete_volume %s', volume['name'])

        freenas_volume = ix_utils.generate_freenas_volume_name(
            volume['name'],
            self.configuration.ixsystems_iqn_prefix)

        if freenas_volume['target']:
            self.common._delete_iscsitarget(freenas_volume['target'])
        if freenas_volume['name']:
            self.common._delete_volume(freenas_volume['name'])

    def create_export(self, context, volume, connector):
        """Driver entry point to get the export info for a new volume."""
        LOG.info('iXsystems Create Export')
        LOG.debug('create_export %s', volume['name'])

        handle = self.common._create_export(volume['name'])
        LOG.info('provider_location: %s', handle)
        return {'provider_location': handle}

    def ensure_export(self, context, volume):
        """Driver entry point to get the export info for an existing volume."""
        LOG.info('iXsystems Ensure Export')
        LOG.debug('ensure_export %s', volume['name'])

        handle = self.common._create_export(volume['name'])
        LOG.info('provider_location: %s', handle)
        return {'provider_location': handle}

    def remove_export(self, context, volume):
        """Driver exntry point to remove an export for a volume.

           we have nothing to do for unexporting.
        """
        pass

    def initialize_connection(self, volume, connector):
        """Driver entry point to attach a volume to an instance."""
        LOG.info('iXsystems Initialise Connection')
        freenas_volume = ix_utils.generate_freenas_volume_name(
            volume['name'],
            self.configuration.ixsystems_iqn_prefix)

        if not freenas_volume['name']:
            # is this snapshot?
            freenas_volume = ix_utils.generate_freenas_snapshot_name(
                volume['name'],
                self.configuration.ixsystems_iqn_prefix)

        properties = {}
        properties['target_discovered'] = False
        properties['target_portal'] = ix_utils.get_iscsi_portal(
            self.configuration.ixsystems_server_hostname,
            self.configuration.ixsystems_server_iscsi_port)
        properties['target_iqn'] = freenas_volume['iqn']
        properties['volume_id'] = volume['id']

        LOG.debug('initialize_connection data: %s', properties)
        return {'driver_volume_type': 'iscsi', 'data': properties}

    def terminate_connection(self, volume, connector, **kwargs):
        """Driver entry point to detach a volume from an instance."""
        pass

    def create_snapshot(self, snapshot):
        """Driver entry point for creating a snapshot."""
        LOG.info('iXsystems Create Snapshot')
        LOG.debug('create_snapshot %s', snapshot['name'])

        freenas_snapshot = ix_utils.generate_freenas_snapshot_name(
            snapshot['name'], self.configuration.ixsystems_iqn_prefix)
        freenas_volume = ix_utils.generate_freenas_volume_name(
            snapshot['volume_name'], self.configuration.ixsystems_iqn_prefix)

        self.common._create_snapshot(freenas_snapshot['name'],
                                     freenas_volume['name'])

    def delete_snapshot(self, snapshot):
        """Driver entry point for deleting a snapshot."""
        LOG.info('iXsystems Delete Snapshot')
        LOG.debug('delete_snapshot %s', snapshot['name'])
        freenas_snapshot = ix_utils.generate_freenas_snapshot_name(
            snapshot['name'],
            self.configuration.ixsystems_iqn_prefix)
        freenas_volume = ix_utils.generate_freenas_volume_name(
            snapshot['volume_name'],
            self.configuration.ixsystems_iqn_prefix)

        self.common._delete_snapshot(freenas_snapshot['name'],
                                     freenas_volume['name'])

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from snapshot."""
        LOG.info('iXsystems Create Volume From Snapshot')
        LOG.info('create_volume_from_snapshot %s', snapshot['name'])

        existing_vol = ix_utils.generate_freenas_volume_name(
            snapshot['volume_name'], self.configuration.ixsystems_iqn_prefix)
        freenas_snapshot = ix_utils.generate_freenas_snapshot_name(
            snapshot['name'], self.configuration.ixsystems_iqn_prefix)
        freenas_volume = ix_utils.generate_freenas_volume_name(
            volume['name'], self.configuration.ixsystems_iqn_prefix)
        freenas_volume['size'] = volume['size']
        freenas_volume['target_size'] = volume['size']

        self.common._create_volume_from_snapshot(freenas_volume['name'],
                                                 freenas_snapshot['name'],
                                                 existing_vol['name'])
        self.common._create_iscsitarget(freenas_volume['target'],
                                        freenas_volume['name'])

    def get_volume_stats(self, refresh=False):
        """Get stats info from volume group / pool."""
        LOG.info('iXsystems Get Volume Status')
        if refresh:
            self.stats = self.common._update_volume_stats()
        LOG.info('get_volume_stats: %s', self.stats)
        return self.stats

    def create_cloned_volume(self, volume, src_vref):
        """Creates a volume from source volume."""
        LOG.info('iXsystems Create Cloned Volume')
        LOG.info('create_cloned_volume: %s', src_vref['id'])

        temp_snapshot = {'volume_name': src_vref['name'],
                         'name': 'name-c%s' % src_vref['id']}

        self.create_snapshot(temp_snapshot)
        self.create_volume_from_snapshot(volume, temp_snapshot)
        # self.delete_snapshot(temp_snapshot)
        # with API v2.0 this causes FreeNAS error
        # "snapshot has dependent clones".  Cannot delete while volume is
        # active.  Instead, added check and deletion of orphaned dependent
        # clones in common._delete_volume()

    def extend_volume(self, volume, new_size):
        """Driver entry point to extend an existing volumes size."""
        LOG.info('iXsystems Extent Volume')
        LOG.info('extend_volume %s', volume['name'])

        freenas_volume = ix_utils.generate_freenas_volume_name(
            volume['name'], self.configuration.ixsystems_iqn_prefix)
        freenas_new_size = new_size

        if volume['size'] != freenas_new_size:
            self.common._extend_volume(freenas_volume['name'],
                                       freenas_new_size)



# Copyright (c) Databunny Pte Ltd
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from driver.freenasapi import FreeNASServer
from driver.freenasapi import FreeNASApiError
import unittest
import json
import os,sys
import urllib.parse
sys.path.append(os.path.abspath(".."))
#sys.path.append(os.path.abspath("./test"))
import configparser
from driver import utils as ix_utils
config = configparser.ConfigParser()
config.read("./src/test/cinder.conf")        
def createvolume(name,size):
    name,size=("test",1)
    freenas = FreeNASServer(config['ixsystems-iscsi']['ixsystems_server_hostname']
                            ,config['ixsystems-iscsi']['ixsystems_server_port']
                                    ,config['ixsystems-iscsi']['ixsystems_login']
                                    ,config['ixsystems-iscsi']['ixsystems_password'])
    params = {}
    params['name'] = config['ixsystems-iscsi']['ixsystems_dataset_path'] + '/' + name
    params['type'] = 'VOLUME'
    params['volsize'] = ix_utils.get_bytes_from_gb(size)
    jparams = json.dumps(params)
    jparams = jparams.encode('utf8')
    request_urn = ('%s') % (FreeNASServer.REST_API_VOLUME)
    ret = freenas.invoke_command(FreeNASServer.CREATE_COMMAND,
                                        request_urn, jparams)

    if ret['status'] != FreeNASServer.STATUS_OK:
        msg = ('Error while creating volume: %s' % ret['response'])
        raise FreeNASApiError('Unexpected error', msg)
    return True
def delvolume(name):
    name,size=("test",1)
    freenas = FreeNASServer(config['ixsystems-iscsi']['ixsystems_server_hostname']
                            ,config['ixsystems-iscsi']['ixsystems_server_port']
                                    ,config['ixsystems-iscsi']['ixsystems_login']
                                    ,config['ixsystems-iscsi']['ixsystems_password'])
    request_urn = ('%s/id/%s%s') % (
        FreeNASServer.REST_API_VOLUME,
        urllib.parse.quote_plus(
            config['ixsystems-iscsi']['ixsystems_dataset_path'] + '/'),
        name)
    ret = freenas.invoke_command(FreeNASServer.DELETE_COMMAND,
                                        request_urn, None)

    if ret['status'] != FreeNASServer.STATUS_OK:
        msg = ('Error while creating volume: %s' % ret['response'])
        raise FreeNASApiError('Unexpected error', msg)
    return True
class testfreenasapi(unittest.TestCase):
    def testcreatevolume(self):
        assert createvolume("test",1)
    def testdelvolume(self):
        assert delvolume("test")    
        
if __name__ == '__main__':
    unittest.main()